module Parser where

import Lexer

data Type = 
    NoType Error |
    TypeInt |
    TypeVoid |
    TypeArray |
    TypePointer deriving (Show, Eq)

data Program = Program [Global] deriving (Show, Eq)
data Global = NoGlobal Error | GlobalFunction Function | GlobalDeclaration Declaration deriving (Show, Eq)
data Function = NoFunction Error | Function (String, Type, Arguments, Block) deriving (Show, Eq)
data Block = NoBlock Error | Block ([Declaration], [Statement]) deriving (Show, Eq)
data Statement =
    NoStatement Error |
    StatementEmpty |
    If (Expression, Statement) |
    IfElse (Expression, Statement, Statement) |
    While (Expression, Statement) |
    InfiniteLoop (Statement) |
    Break |
    StatementBlock Block |
    StatementExpression (Expression)
    deriving (Show, Eq)
data Variable =
    NoVariable Error |
    Variable (Type, String, Int) deriving (Show, Eq)
data Expression =
    Number Int |
    VariableValue String |
    PreOperation1 (String, Expression) |
    PostOperation1 (String, Expression) |
    Operation2 (String, Expression, Expression) |
    Call (String, Expression) |
    ArrayIndex (String, Expression) |
    ExpressionEmpty |
    NoExpression Error
    deriving (Show, Eq)
data Arguments = Arguments [Argument] | ArgumentsVoid deriving (Show, Eq)
data Argument = NoArgument Error | Argument (Type, Variable) deriving (Show, Eq)
data Declaration = NoDeclaration Error |
                   Declaration [(Variable, Initialization)] deriving (Show, Eq)

data Initialization =
    NoInitialization Error |
    InitializationNone |
    InitializationInt Int |
    InitializationIntArray [Int]
    deriving (Show, Eq)

data Error = Error [String] | NoError deriving (Show, Eq)

globalError (NoGlobal e) = e
globalError _ = NoError
functionError (NoFunction e) = e
functionError _ = NoError
expressionError (NoExpression e) = e
expressionError _ = NoError
variableError (NoVariable e) = e
variableError _ = NoError
argumentError (NoArgument e) = e
argumentError _ = NoError
declarationError (NoDeclaration e) = e
declarationError _ = NoError
statementError (NoStatement e) = e
statementError _ = NoError
typeError (NoType e) = e
typeError _ = NoError
declarationVariableError (NoVariable e, _) = e
declarationVariableError (_, NoInitialization e) = e
declarationVariableError _ = NoError
           
tree = fst
remainTokens = snd

parse tokens =
    if remainTokens program == [] then
        tree program
    else
        parseErrorStop "invalid syntax" (remainTokens program)
    where
      program = parseProgram tokens

parseError t message tokens = (t (Error [parseErrorMessage message tokens]), tokens)
parseErrorStop message tokens = error (parseErrorMessage message tokens)
parseErrorMessage message tokens = ("parse error: " ++ message ++ " at " ++ errorTokenShow tokens)
errorTokenShow:: [Token] -> String
errorTokenShow (token:tokens) = show token
errorTokenShow _ = "?"

eatCheckToken t (token:tokens) | token == t = tokens
eatCheckToken t tokens = parseErrorStop ("can't found " ++ show t) tokens

eatCheckTokens (t:ts) tokens = eatCheckTokens ts (eatCheckToken t tokens)
eatCheckTokens [] tokens = tokens

checkNextToken p [] = False
checkNextToken p (token:tokens) = p token

parseProgram tokens =
    (Program (tree globals), remainTokens globals)
    where
      globals = parseGlobals tokens
                       
parseGlobals = parseSequence parseGlobal globalError

parseGlobal tokens = 
    if functionError (tree tryFunction) == NoError
    then (GlobalFunction (tree tryFunction), remainTokens tryFunction)
    else if declarationError (tree tryDeclaration) == NoError
         then (GlobalDeclaration (tree tryDeclaration), remainTokens tryDeclaration)
         else parseError NoGlobal "invalid global definition" tokens
    where
      tryFunction = parseFunction tokens
      tryDeclaration = parseDeclaration tokens

parseFunction tokens =
    case ftypeRemain of
      ((TokenIdentifier fname):(TokenSymbol "("):rtokens) ->
          let 
              args = parseArguments rtokens
              argsRemain = eatCheckToken (TokenSymbol ")") (remainTokens args)
              block = parseBlock argsRemain
          in
            (Function (fname, tree ftype, tree args, tree block), remainTokens block)
      _ -> parseError NoFunction "invalid function" tokens
    where
      ftype = parseType tokens
      ftypeRemain = remainTokens ftype

parseBlock ((TokenSymbol "{"):tokens) = 
    (Block (tree declarations, tree statements), eatCheckToken (TokenSymbol "}") (remainTokens statements))
    where
      declarations = parseDeclarations tokens
      statements = parseStatements (remainTokens declarations)
parseBlock tokens = parseError NoBlock "invalid block" tokens

parseStatements = parseSequence parseStatement statementError

parseStatement ((TokenIdentifier "break"):(TokenSymbol ";"):tokens) =
    (Break, tokens)

parseStatement ((TokenIdentifier "while"):(TokenSymbol "("):tokens) =
    (While (tree cond, tree cont), remainTokens cont)
    where
      cond = parseExpression tokens
      cont = parseStatement (eatCheckToken (TokenSymbol ")") (remainTokens cond))

parseStatement ((TokenIdentifier "for"):(TokenSymbol "("):tokens) =
    (StatementBlock
     (Block
      ([],
       [StatementExpression (tree e1),
        While (tree e2, StatementBlock (Block ([], [tree cont, StatementExpression (tree e3)])))]
      )
     ),
     remainTokens cont)
    where
      e1 = if checkNextToken (==(TokenSymbol ";")) tokens
           then (Number 1, tokens)
           else parseExpression tokens
      e1tokens = (eatCheckToken (TokenSymbol ";") (remainTokens e1))

      e2 = if checkNextToken (==(TokenSymbol ";")) e1tokens
           then (Number 1, e1tokens)
           else parseExpression e1tokens
      e2tokens = (eatCheckToken (TokenSymbol ";") (remainTokens e2))

      e3 = if checkNextToken (==(TokenSymbol ")")) e2tokens
           then (Number 1, e2tokens)
           else parseExpression e2tokens
      e3tokens = (eatCheckToken (TokenSymbol ")") (remainTokens e3))

      cont = parseStatement e3tokens

parseStatement ((TokenIdentifier "if"):(TokenSymbol "("):tokens) =
    if head thenRemain == (TokenIdentifier "else") then
        let
            elseStatemant = parseStatement (tail thenRemain)
        in
          (IfElse (tree cond, tree thenStatemant, tree elseStatemant), remainTokens elseStatemant)
    else
        (If (tree cond, tree thenStatemant), thenRemain)
    where
      cond = parseExpression tokens
      thenStatemant = parseStatement (eatCheckToken (TokenSymbol ")") (remainTokens cond))
      thenRemain = (remainTokens thenStatemant)

parseStatement atokens@((TokenSymbol "{"):tokens) =
    (StatementBlock (tree block), remainTokens block)
    where
      block = parseBlock atokens

parseStatement ((TokenSymbol ";"):tokens) =
    (StatementEmpty, tokens)

parseStatement tokens =
    if (expressionError (tree expression)) == NoError
    then
        (StatementExpression (tree expression), eatCheckToken (TokenSymbol ";") (remainTokens expression))
    else
        parseError NoStatement "invalid statement" tokens
    where
      expression = parseExpression tokens

parseDeclarations = parseSequence parseDeclaration declarationError

parseDeclaration tokens =
    if typeError (tree ts) == NoError then
        (Declaration (tree vs),
         eatCheckToken (TokenSymbol ";") (remainTokens vs))
    else
        parseError NoDeclaration "invalid declaration" tokens
    where
      ts = parseType tokens
      vs = parseDeclarationVariables (tree ts) (remainTokens ts)

parseDeclarationVariables vtype =
    parseEnumeration (TokenSymbol ",") (parseDeclarationVariable vtype) declarationVariableError

parseDeclarationVariable vtype tokens =
    ((tree v, tree ini), remainTokens ini)
    where
      v = parseVariable vtype tokens
      ini = parseInitialization (remainTokens v)

parseInitialization ((TokenSymbol "="):(TokenSymbol "{"):tokens) =
    (InitializationIntArray (map (\(Number i) -> i) (tree elements)),
     eatCheckToken (TokenSymbol "}") (remainTokens elements))
    where
      elements = parseEnumeration (TokenSymbol ",") parseConstantExpression expressionError tokens
parseInitialization ((TokenSymbol "="):tokens) =
    (InitializationInt ((\(Number x) -> x) (tree constant)), remainTokens constant)
    where
      constant = parseConstantExpression tokens
parseInitialization tokens =
    (InitializationNone, tokens)

parseConstantExpression ((TokenSymbol "-"):(TokenNumber i):tokens) = (Number (-i), tokens)
parseConstantExpression ((TokenNumber i):tokens) = (Number i, tokens)
parseConstantExpression tokens = parseError NoExpression "invalid constant " tokens

parseVariables vtype = parseEnumeration (TokenSymbol ",") (parseVariable vtype) variableError

parseVariable vtype ((TokenIdentifier vname):(TokenSymbol "["):(TokenNumber i):(TokenSymbol "]"):tokens) = 
    (Variable (TypeArray, vname, i), tokens)
parseVariable vtype ((TokenIdentifier vname):(TokenSymbol "["):(TokenSymbol "]"):tokens) = 
    (Variable (TypeArray, vname, 0), tokens)
parseVariable vtype ((TokenIdentifier vname):tokens) = (Variable (vtype, vname, 1), tokens)
parseVariable vtype ((TokenSymbol "*"):tokens) = 
    case fst e of
      (Variable (t, vn, _)) ->
          (Variable (TypePointer, vn, 1), remainTokens e)
      _ -> parseError NoVariable "invalid variable" tokens
    where
      e = parseVariable vtype tokens
parseVariable vtype tokens = parseError NoVariable "invalid variable" tokens

parseArguments ((TokenIdentifier "void"):(TokenSymbol ")"):tokens) = 
    (ArgumentsVoid, (TokenSymbol ")"):tokens)
parseArguments tokens =
    (Arguments (tree arguments), remainTokens arguments)
    where
      arguments = parseEnumeration (TokenSymbol ",") parseArgument argumentError tokens

parseArgument tokens =
    if typeError (tree ts) == NoError && variableError (tree vs) == NoError then
        (Argument (tree ts, tree vs), (remainTokens vs))
    else
        parseError NoArgument "invalid argument" tokens
    where
      ts = parseType tokens
      vs = parseVariable (tree ts) (remainTokens ts)

parseExpression = parseOperation2LeftExpression1

parseOperation2LeftExpression1 =
    foldl parseOperation2Left parseOperation2RightExpression
              [(== ",")]

parseOperation2RightExpression =
    foldl parseOperation2Right parseOperation2LeftExpression2
              [(\x -> any (==x) ["=", "+=", "-=", ">>="])]

parseOperation2LeftExpression2 =
    foldl parseOperation2Left parsePreOperation1RightExpression
              [(\x -> any (==x) ["*", "/"]),
               (\x -> any (==x) ["+", "-"]),
               (\x -> any (==x) [">>"]),
               (\x -> any (==x) ["<", "<=", ">", ">="]),
               (\x -> any (==x) ["==", "!="]),
               (== "&&"),
               (== "||")]

parsePreOperation1RightExpression =
    foldl parsePreOperation1Right parsePostOperation1LeftExpression
              [(\x -> any (==x) ["++", "--", "+", "-", "&", "!", "*", "~"])]

parsePostOperation1LeftExpression =
    foldl parsePostOperation1Left parseFactor
              [(\x -> any (==x) ["++", "--"])]

parseFactor ((TokenNumber i):tokens) = (Number i, tokens)
parseFactor ((TokenIdentifier vname):(TokenSymbol "["):tokens) =
    (ArrayIndex (vname, tree expression), eatCheckToken (TokenSymbol "]") (remainTokens expression))
    where
      expression = parseExpression tokens
parseFactor ((TokenIdentifier fname):(TokenSymbol "("):(TokenSymbol ")"):tokens) =
    (Call (fname, ExpressionEmpty), tokens);
parseFactor ((TokenIdentifier fname):(TokenSymbol "("):tokens) = 
    (Call (fname, tree expression), eatCheckToken (TokenSymbol ")") (remainTokens expression))
    where
      expression = parseExpression tokens

parseFactor ((TokenIdentifier v):tokens) = (VariableValue v, tokens)
parseFactor ((TokenSymbol "("):tokens) =
    if head r /= (TokenSymbol ")") then
        parseError NoExpression "invalid factor" r
    else
        (tree t, tail r)
    where
      t = parseExpression tokens
      r = remainTokens t
parseFactor tokens = parseError NoExpression "invalid factor" tokens

parseType (TokenIdentifier "int":tokens) = (TypeInt, tokens)
parseType (TokenIdentifier "void":tokens) = (TypeVoid, tokens)
parseType tokens = parseError NoType "invalid type" tokens

parseSequence pf errorGetter tokens =
    if (errorGetter at) == NoError then
        let next = parseSequence pf errorGetter r
            nextTree = tree next
            nextTokens = remainTokens next
        in
          (at:nextTree, nextTokens)
    else
        ([], tokens)
    where
      a = pf tokens
      at = tree a
      r = remainTokens a

parseEnumeration delimiterToken pf errorGetter tokens =
    if (errorGetter at) == NoError then
        case r of
          (t:rr) | delimiterToken == t -> 
               let next = parseEnumeration delimiterToken pf errorGetter rr
                   nextTree = tree next
                   nextTokens = remainTokens next
               in
                 (at:nextTree, nextTokens)
          _ -> ([at], r)
    else
        ([], tokens)
    where
      a = pf tokens
      at = tree a
      r = remainTokens a

parsePostOperation1Left p1 isSymbol tokens =
    parsePostOperation1Left' p1 isSymbol (remainTokens t) (tree t)
    where
      t = p1 tokens
parsePostOperation1Left' p1 isSymbol atokens@((TokenSymbol s):tokens) t =
    if isSymbol s
    then
        parsePostOperation1Left' p1 isSymbol tokens (PostOperation1 (s, t))
    else
        (t, atokens)
parsePostOperation1Left' p1 isSymbol tokens t = (t, tokens)

parsePreOperation1Right p1 isSymbol ((TokenSymbol s):tokens)
    | isSymbol s =
        (PreOperation1 (s, tree t), remainTokens t)
        where
          t = parsePreOperation1Right p1 isSymbol tokens
parsePreOperation1Right p1 isSymbol tokens = p1 tokens

parseOperation2Left p1 isSymbol tokens =
    parseOperation2Left' p1 isSymbol (remainTokens t) (tree t)
    where
      t = p1 tokens
parseOperation2Left' p1 isSymbol atokens@((TokenSymbol s):tokens) t = 
    if isSymbol s then
        parseOperation2Left' p1 isSymbol (remainTokens t1) (Operation2 (s, t, tree t1))
    else (t, atokens)
    where
      t1 = p1 tokens
parseOperation2Left' p1 isSymbol tokens t = (t, tokens)

parseOperation2Right p1 isSymbol tokens =
    parseOperation2Right' p1 isSymbol (remainTokens t) (tree t)
    where
      t = p1 tokens
parseOperation2Right' p1 isSymbol atokens@((TokenSymbol s):tokens) t = 
    if isSymbol s then
        (Operation2 (s, t, tree t1), (remainTokens t1))
    else (t, atokens)
    where t1 = parseOperation2Right p1 isSymbol tokens
parseOperation2Right' p1 isSymbol tokens t = (t, tokens)
